"""
    antidox.directives
    ~~~~~~~~~~~~~~~~~~

    C domain directives for documenting Doxygen elements. These directives try
    to take advantage of the work done by Doxygen to avoid having to parse C
    source.

"""

import re

from lxml import etree as ET
from docutils.parsers.rst import Directive, directives
from docutils.nodes import Text, Structural, literal, Element, paragraph
from sphinx.util import logging
from sphinx.util.nodes import split_explicit_title, nested_parse_with_titles
from sphinx.locale import _ as _locale
from sphinx.domains import Domain
from sphinx import addnodes
import sphinx.errors

from . import doxy
from .xtransform import get_stylesheet
from .nodes import (nodeclass_from_tag, PlaceHolder, DeferredPlaceholder,
                    FakeRoot)
from .collector import DoxyCollector

__author__ = "Juan I Carrano"
__copyright__ = "Copyright 2018, Freie Universität Berlin"


logger = logging.getLogger(__name__)


class InvalidEntity(sphinx.errors.SphinxError):
    pass


ENTITY_RE = re.compile(r"(?:!(?P<refid>[\w-]+))|(?:(?P<kind>\w+)?\[(?P<name>[-\w]+)\])|(?P<target>[^[]\S*)")
"""Regular expression for references to entities.

Catches either target strings (a/b.h::c), kind[name] strings or refid strings
(prefixed with an exclamation mark "!").

Capture groups:

``refid``:
    If the string starts with "!", then returns the rest of it, unmodified.

``target``
    The target string if the reference is a antidox.doxy.Target-compatible
    target spec, else None.

``kind``
    None if kind was not specified or if the string is a target.

``name``
    The name if the string is kind[name] or [name] (i.e. not a target).
"""

def _get_current_scope(env):
    """Get the refid that is being currently documented. Also, create the scope
    stack it it does not exist."""
    scope_stack = env.ref_context.setdefault('doxy:refid', [])

    try:
        scope = scope_stack[-1]
    except IndexError:
        scope = 0

    return scope


def resolve_refstr(env, ref_str):
    """Transform a reference string (see :py:data:`ENTITY_RE`) into a RefId.
    If ref_str is already a refid, it is still validated.

    Returns
    -------

    ref: RefId for the element
    ref_spec: re.Match object, the result of parsing ref_str.
    """

    ref_spec = ENTITY_RE.fullmatch(ref_str)

    if ref_spec is None:
        raise InvalidEntity("Cannot parse entity: %s" % ref_str)

    scope = _get_current_scope(env)

    db = env.antidox_db

    target = ref_spec['target']
    refid_s = ref_spec['refid']

    if target:
        ref = db.resolve_target(target, scope)
    elif refid_s:
        # just validate that the reference is valid
        ref = doxy.RefId(refid_s)
        db.get(ref)
    else:
        kind_s = ref_spec['kind']
        ref = db.resolve_name(kind_s and doxy.Kind.from_attr(kind_s),
                              ref_spec['name'], scope)

    return ref, ref_spec


class _Universal:
    """Container containing everything."""
    def __contains__(self, k):
        """Returns True always because this container contains everything."""
        return True


_Universe = _Universal()
"""Singleton universal object"""


def _empty_to_universe(s):
    return (s or _Universe) if s is not None else ()


_DEFAULT_CHILDREN_KINDS = (doxy.Kind.ENUM, doxy.Kind.STRUCT)


def _add_default_options_(inherited, kind, this_kind):
    option_copy = inherited.copy()

    if kind == this_kind:
        option_copy.update({'no-children': ''})
    if kind in _DEFAULT_CHILDREN_KINDS:
        option_copy.update({'children': ''})

    return option_copy


def default_inclusion_policy(app, this, options):
    """Default behavior used when no callback handles the
    ``antidox-include-default`` event.

    If ``no-children`` is empty, it is taken to mean "exclude everything". If
    ``children`` is empty it is taken include everything (with exceptions).

    The exceptions are that children with the same kind as the parent won't be
    included unless explicitly named and if they are included they will have
    ``no-children`` set.

    Flag options are inherited by default. Some kinds ("struct" and "enum", will
    have "children" set by default.)
    """
    db = app.env.antidox_db

    no_children = _empty_to_universe(options.get('no-children'))
    yes_children = _empty_to_universe(options.get('children'))

    if yes_children and no_children is not _Universe:
        all_child_members, all_child_compounds = db.find_children(this)

        this_kind = db.get(this)['kind']

        # By default, inherit all flags
        inherited_options = {k: v for k, v in options.items()
                             if k in DoxyExtractor._flag_parameters}
        inherited_nochildren = inherited_options.copy().update(
            {'no-children': ''})

        def _member_accept(name, kind):
            return ((kind != this_kind) if yes_children is _Universe
                    else (name in yes_children))

        inclusion_list = [(ref, _add_default_options_(inherited_options, kind,
                                                      this_kind))
                          for ref, name, kind in all_child_members
                          if name in yes_children
                          and name not in no_children]

        inclusion_list.extend(
            (ref, _add_default_options_(inherited_options, kind, this_kind))
            for ref, name, kind in all_child_compounds
            if _member_accept(name, kind)
            and name not in no_children)

        return inclusion_list


def string_list(argument):
    """Parse a list of strings. If no argument is given, return an empty
    list."""
    return argument.split() if argument is not None else []


_STR2BOOL = {"false": False, "true": True}


class DoxyExtractor(Directive):
    """
    Auto-document any doxygen entity:

    - C language elements are specified as a target string (as in
      antidox.doxy.Target.
    - Arbitrary doxygen entities can be specified as kind[name], where kind is
      optional (as long as there is no ambiguity.

    Options
    -------

    hidedef: hide macro definition.
    hideloc: hide the location of the definition.
    hidedoc: hide doxygen's documentation.
    """

    has_content = True
    required_arguments = 1
    optional_arguments = 0

    option_spec = {
        'noindex': directives.flag,
        'hidedef': directives.flag,
        'hideloc': directives.flag,
        'hidedoc': directives.flag,
        'children': string_list,
        'no-children': string_list,
    }

    _flag_parameters = [opt for opt, typ in option_spec.items()
                        if typ == directives.flag]

    def add_target_and_index(self, name, sig, signode):
        # type: (Any, unicode, addnodes.desc_signature) -> None
        """
        Add cross-reference IDs and entries to self.indexnode, if applicable.

        Parameters
        ----------

        name: refid for the object.
        sig: Signature. It is parsed as a doxy.Target string.
        """
        # FIXME: what is this supposed to do?

        return  # do nothing by default

    @property
    def env(self):
        """Shortcut to get this document's environment."""
        return self.state.document.settings.env

    @property
    def db(self):
        """Get the DoxyDB object."""
        return self.env.antidox_db

    @staticmethod
    def _attr_to_obj(attr_string):
        """Try to convert a string to a python object"""
        if attr_string.isdigit():
            return int(attr_string)
        else:
            return _STR2BOOL.get(attr_string, attr_string)

    def _etree_to_sphinx(self, etree):
        """Convert an element tree to sphinx nodes.

        A text node with a antidox:l attribute will be translated using sphinx
        locale features.

        Return
        ------

        nodes: List of sphinx nodes
        special: dictionary if special placeholder nodes.
            Currently defined elements are:
            antidox_children
                placeholder that should be replaced by this element's children.
        """
        curr_element = FakeRoot()
        root = curr_element

        special = {}

        if etree.getroot() is None:
            logger.warn("Template produced no elements for %s",
                        self.arguments[0])
            return [], special

        for action, elem in ET.iterwalk(etree, events=("start", "end")):
            if action == "start":
                nclass = nodeclass_from_tag(elem.tag)

                arg = elem.text if issubclass(nclass, Text) else ''

                # automatically handle list attributes
                list_attributes = getattr(nclass, "list_attributes", ())
                filtered_attrs = {k: (v.split("|")
                                      if k in list_attributes
                                      else self._attr_to_obj(v))
                                  for (k, v) in elem.attrib.items()}

                node = nclass(arg, **filtered_attrs)
                if not isinstance(node, Text) and elem.text:
                    node += Text(elem.text, elem.text)

                curr_element.append(node)

                # FIXME: this smells hacky
                if isinstance(node, Structural) and node['ids']:
                    self.state.document.note_explicit_target(node)

                curr_element = node
            else:
                if isinstance(curr_element, PlaceHolder):
                    curr_element.replace_placeholder(self.lineno, self.state,
                                                     self.state_machine)

                if isinstance(curr_element, DeferredPlaceholder):
                    special[curr_element.tagname] = curr_element

                curr_element = curr_element.parent

                if elem.tail:
                    curr_element.append(Text(elem.tail, elem.tail))

        return root.children, special

    def _process_content(self, nodes, special):
        """Parse the directive content and place it in the appropiate nodes.

        Parameters
        ----------
        nodes, special: the result of run_reference.

        Returns
        -------
        nodes: new list of nodes
        """
        uccontent = special.get("antidox_usercontent")

        assert nodes if uccontent else True

        # If there is a antidox_usercontent node, it should be REPLACED by
        # the user content, but if not, we must find a default placement node
        # and NEST the user content UNDER that node.
        if uccontent:
            # make a temporary container
            content_container = Element()
        elif not nodes:
            # This may give weird unexpected results when children are added.
            content_container = paragraph()
            nodes = [content_container]
        else:
            child_index = nodes[0].first_child_matching_class(
                addnodes.desc_content)
            content_container = (nodes[0][child_index]
                                 if child_index is not None else nodes[-1])

        if self.content:
            nested_parse_with_titles(self.state, self.content,
                                     content_container)

        if uccontent:
            # handle the case where antidox_usercontent is at the top level
            if uccontent.parent:
                uccontent.replace_self(content_container.children)
            else:
                uc_index = nodes.index(uccontent)
                nodes[uc_index:uc_index+1] = content_container.children
            assert nodes if self.content else True

        return nodes

    def _options_to_params(self):
        return {k: 'true()' if k in self.options else 'false()'
                for k in self._flag_parameters}

    def run_reference(self, ref):
        """Convert the doxygen XML of a reference into Sphinx nodes.

        Parameters
        ----------
        ref: a antidox.Doxy.RefId (or string)

        Returns
        -------
        nodes: List of sphinx nodes.
        special: a dictionary of special nodes (subclasses of DeferredPlaceholder)
        """

        element_tree = self.db.get_tree(ref)

        my_domain = self.env.domains['doxy']

        rst_etree = my_domain.stylesheet(element_tree,
                                         **self._options_to_params())
        nodes, special = self._etree_to_sphinx(rst_etree)

        style_fn = my_domain.stylesheet_filename
        if style_fn:
            self.env.note_dependency(style_fn)

        DoxyCollector.note_dependency(self.env)

        return nodes, special

    def run(self):
        arg0 = self.arguments[0]

        if isinstance(arg0, doxy.RefId):
            ref = arg0
        else:
            ref = resolve_refstr(self.env, arg0)[0]

        context_stack = self.env.ref_context['doxy:refid']
        context_stack.append(ref)

        try:
            nodes, special = self.run_reference(ref)

            nodes = self._process_content(nodes, special)

            ev_args = (ref, self.options)
            ev_result = self.env.app.emit_firstresult(
                                "antidox-include-default", *ev_args)
            inclusion_list = (ev_result if ev_result is not None else
                              (default_inclusion_policy(self.env.app, *ev_args)
                               or [])
                              )

            self.env.app.emit_firstresult(
                                "antidox-include-children",
                                ref, self.options, inclusion_list)

            this_directive = type(self)

            nodes_to_insert = (internal_node
                               for refid, options in inclusion_list
                               for internal_node in this_directive(
                                   'doxy:c', [refid], options, [], self.lineno,
                                   0, "", self.state, self.state_machine).run())
        finally:
            context_stack.pop()

        # just in case the elements produced not output, it should not be an
        # error.
        if "antidox_children" in special:
            # replace_self needs an indexable sequence
            special["antidox_children"].replace_self(list(nodes_to_insert))
        else:
            (nodes[-1] if nodes else nodes).extend(nodes_to_insert)

        return nodes


def target_role(typ, rawtext, text, lineno, inliner, options={}, content=[]):
    """Create a cross reference for a doxygen object, given a human-readable
    target.

    The target is interpreted by resolve_refstr. It can be prefixed by ``~``,
    which causes target-strings to drop the path.
    """

    env = inliner.document.settings.env
    db = env.antidox_db

    is_explicit, title, _target = split_explicit_title(text.strip())

    title = title.strip()
    _target = _target.strip()

    remove_path = _target.startswith("~")

    target = _target[1:] if remove_path else _target

    try:
        ref, match = resolve_refstr(env, target)
    except doxy.RefError as e:
        msg = inliner.reporter.error(e.args[0], line=lineno)
        prb = inliner.problematic(rawtext, rawtext, msg)
        return [prb], [msg]

    try:
        reftype = db.guess_desctype(ref)
    except ValueError:
        # if there is no type, it means it is a construct that does not fit in
        # the C domain and therefore it will be a cross reference in std.
        refdomain = 'std'
        reftype = 'ref'
        innernode = Text
    else:
        refdomain = 'c'
        innernode = literal

    node = addnodes.pending_xref(rawsource=rawtext, reftarget=str(ref),
                                 refdomain=refdomain, reftype=reftype,
                                 refexplicit=is_explicit,
                                 refdoc=env.docname,
                                 refwarn=True)
    if not is_explicit:
        # FIXME: make this fomatting customizable
        if match["name"]:
            linktext = match["name"]
        else:
            doxy_target = (doxy.Target(match["target"]) if match["target"]
                           else db.refid_to_target(match["refid"]))
            linktext = doxy_target.name if remove_path else str(doxy_target)

        if reftype == "function":
            linktext += "()"

        if match["kind"]:
            linktext += " ({})".format(match["kind"])

        node += innernode(linktext, linktext)
    else:
        node += Text(title, title)

    DoxyCollector.note_dependency(env)

    return [node], []


class DoxyDomain(Domain):
    """Domain for Doxygen-related directives and roles.

    The cross reference data is stored in the C domain. The only reason this
    domain exists is to serve as a container for template and other data that
    should be shared but not be saved with the environment.

    Attributes
    ----------

    DoxyDomain.stylesheet_filename: File name of the XSL stylesheet. If None
        or empty, it means the embedded stylesheet that comes with this
        extension is being used.
    DoxyDomain.stylesheet: An lxml.etree.XSLT object to be used as a stylesheet
        for converting doxygen xml into reST nodes.
    """
    name = 'doxy'
    label = "Doxygen-documented entities"

    directives = {'c': DoxyExtractor}
    roles = {'r': target_role}

    def __init__(self, env):
        super().__init__(env)

        self.stylesheet_filename = env.app.config.antidox_xml_stylesheet

        env.app.connect("antidox-db-loaded", self._load_stylesheet)

    def _load_stylesheet(self, app, db):
        self.stylesheet = get_stylesheet(self.stylesheet_filename,
                                         locale_fn=_locale,
                                         doxy_db=db)

    def merge_domaindata(self, docnames, otherdata):
        """Nothing to do here."""
        pass



"""
    antidox.nodes
    ~~~~~~~~~~~~~

    Special reST nodes used by the XML engine. These nodes never make it to the
    final output. They are replaced by the doxy:c directive.
"""

import abc

from docutils import nodes as _nodes
from docutils.parsers.rst import directives, DirectiveError
from sphinx import addnodes

__author__ = "Juan I Carrano"
__copyright__ = "Copyright 2018, Freie Universität Berlin"


class PseudoElementMeta(abc.ABCMeta):
    """Metaclass for all elements which appear in the output of the XSLT filter
    but are not actual reST elements.

    PseudoElementMeta.tag_map keeps a registry of tag name -> class.
    """
    tag_map = {}

    def __new__(mcls, name, bases, namespace, **kwargs):
        _name = namespace.pop("_name", None)
        if _name:
            namespace.setdefault("tagname",  "antidox_{}".format(_name))
            namespace.setdefault("TAG_NAME", "{{antidox}}{}".format(_name))

        cls = super().__new__(mcls, name, bases, namespace, **kwargs)

        if (not isinstance(cls.TAG_NAME, property)
            or not cls.TAG_NAME.__isabstractmethod__):
            mcls.tag_map[cls.TAG_NAME] = cls

        return cls


class PseudoElement(metaclass=PseudoElementMeta):
    """Base class for elements that get replaced by the "c" directive.

    The metaclass ensures that if ``_name`` is set (to the tag name without
    namespace) TAG_NAME and tagname will be set automatically.
    """
    @property
    @abc.abstractmethod
    def TAG_NAME(self):
        """Tag name for the XML transformer. Should be of the form
        {antidox}name to be correctly namespaced (it will result in a tag
        named <antidox:name>."""

    @property
    @abc.abstractmethod
    def tagname(self):
        """Tag name for docutils."""


class PlaceHolder(PseudoElement):
    """Placeholder elements must be replaced before exiting when traversing the
    element tree.

    They must implement a method `e.replace_placeholder(state)` or a
    `run_directive`
    """
    def replace_placeholder(self, *args):
        """Run the directive and replace this node by the directive's output."""
        new = self.run_directive(*args)

        self.replace_self(new)


class FakeRoot(PlaceHolder, _nodes.Element):
    """Sometimes it is desired to emit several nodes (a list) from a template.
    This implementation transforms each doxygen entity as a root node, and
    because well formed XML can only have one root node, this would be
    impossible.

    """
    _name = "fakeroot"

    def replace_placeholder(self, *args):
        """Replace this element by it's children."""
        self.replace_self(self.children)


class DeferredPlaceholder(PseudoElement, _nodes.Element):
    """Base class for placeholders that are replaced by content that depends on
    the directive contents and options."""


class Children(DeferredPlaceholder):
    """This element gets replaced by the documentation of the current entity's
    children."""
    _name = "children"


class UserContent(DeferredPlaceholder):
    """Placeholder for the content given in the directive's body."""
    _name = "usercontent"


class Index(PlaceHolder, _nodes.Inline, _nodes.TextElement):
    """Add a cross-referenceable index entry to the parent of this element.
    """
    _name = "index"
    # ~ ``(entrytype, entryname,
    # ~ target, ignored, key)``.

    def guess_objtype(self):
        """Try to find an objtype tag in an ancestor of this node."""

        ancestor = self.parent

        while True:
            if ancestor is None:
                raise ValueError("objtype not found")

            try:
                return ancestor['objtype']
            except KeyError:
                pass

            ancestor = ancestor.parent

        raise ValueError("objtype not found")

    def run_directive(self, *args, **kwargs):
        key = self.get("key", None)
        return [addnodes.index(entries=[("single", name, id_, '', key)])
                for id_, name in zip(self.parent['ids'], self.parent['names'])]

    def replace_placeholder(self, lineno, state, state_machine):
        state.document.note_explicit_target(self.parent)

        env = state.document.settings.env

        for domain, _, sref in (s.partition(".") for s in self.parent['ids']):
            inv = env.domaindata[domain]['objects']
            if sref in inv:
                state_machine.reporter.warning(
                    'duplicate %s object description of %s, ' % (domain, sref) +
                    'other instance in ' + env.doc2path(inv[sref][0]),
                    line=lineno)  # FIXME
            inv[sref] = (env.docname, self.guess_objtype())

        return super().replace_placeholder(lineno, state, state_machine)


class Interpreted(PlaceHolder, _nodes.Element):
    _name = "interpreted"

    def run_directive(self, lineno, state, state_machine):
        text = self[0].astext()

        # If we did not parse anything before, language is not set and there is
        # an exception. I don't know if this is the proper way to fix it, but
        # I know no better. See the definition of Inliner.interpreted in
        # docutils to learn more.
        if not hasattr(state.inliner, "language"):
            state.inliner.parse("", lineno, state_machine.memo, None)

        nodes, messages = state.inliner.interpreted('', text,
                                                    self['role'], lineno)

        # fixme: do somethin with messages
        return nodes


class DirectiveArg(PseudoElement, _nodes.Text):
    _name = "directive-argument"


class DirectiveContent(PseudoElement, _nodes.Text):
    _name = "directive-content"


class DirectivePlaceholder(PlaceHolder, _nodes.Element):
    """A directive is specified using <antidox:directive>,
    <antidox:directive-argument>, <antidox:directive-content> tags.

    Each <antidox:directive> can contain zero or more <antidox:argument> and
    an optional content.

    <antidox:directive> Attributes:

    - antidox:name    Name of the directive
    - other tags: Intgerpreted as directive options.

    <antidox:directive-argument> Arguments to the directive.
    <antidox:directive-content> Text contents of the directive.
    """
    _name = "directive"

    def run_directive(self, lineno, state, state_machine):
        """Execute the directive generate a list of nodes."""

        name = self["{antidox}name"]

        directive_class, messages = directives.directive(
                                    name, state.memo.language, state.document)

        raw_options = self.attlist()
        options = {k: directive_class.option_spec[k](v) for k, v in raw_options
                   if not k.startswith("{antidox}")}

        arguments = [n.astext() for n in self.children
                     if n.tagname == "antidox_directive-argument"]
        content = [n.astext() for n in self.children
                   if n.tagname == "antidox_directive-content"]

        # what about this?
        # content_offset = 0
        # block_text = ''

        directive_instance = directive_class(
                            name, arguments, options, content, lineno,
                            0, "", state, state_machine)

        try:
            result = directive_instance.run()
            result += messages
        except DirectiveError as error:
            msg_node = state.reporter.system_message(error.level, error.msg,
                                                     line=lineno)
            result = [msg_node]

        return result


def nodeclass_from_tag(tag):
    """Get the class for a reST node given its tag name.
    This searches in antidox's pseudo-elements, docutils' nodes and sphinx'
    addnodes.
    """
    if tag in PseudoElementMeta.tag_map:
        return PseudoElementMeta.tag_map[tag]

    try:
        return getattr(addnodes, tag)
    except AttributeError:
        return getattr(_nodes, tag)



"""
    antidox
    ~~~~~~~

    Access Doxygen documentation from within Sphinx. Provides autodoc-like
    functionality.
"""

import os

from sphinx.util import logging

from . import doxy
from . import directives
from .collector import DoxyCollector

__author__ = "Juan I Carrano"
__copyright__ = "Copyright 2018, Freie Universität Berlin"


logger = logging.getLogger(__name__)


def load_db(app):
    cfgdir = app.config.antidox_doxy_xml_dir

    cfgdir_time = os.path.getmtime(cfgdir) if os.path.exists(cfgdir) else None

    logger.debug("Doxy XML last modified: %s", cfgdir_time)

    env = app.env

    if (not hasattr(env, "antidox_db")
        or cfgdir_time is None
        or not hasattr(env, "antidox_db_date")
        or env.antidox_db_date < cfgdir_time):

        logger.info("(Re-)Reading Doxygen DB")
        env.antidox_db = doxy.DoxyDB(cfgdir)
        env.antidox_db_date = cfgdir_time

    app.emit("antidox-db-loaded", env.antidox_db)


def setup(app):
    app.add_config_value("antidox_doxy_xml_dir", "", 'env')
    app.add_config_value("antidox_xml_stylesheet", "", 'env')
    app.add_event("antidox-include-default")
    app.add_event("antidox-include-children")
    app.add_event("antidox-db-loaded")

    # TODO: provide support for multiple Doxygen projects
    app.connect("builder-inited", load_db)
    app.add_env_collector(DoxyCollector)

    # app.add_directive('doxy', directives.CAuto)
    # roles.register_canonical_role('doxyt', directives.target_role)
    app.add_domain(directives.DoxyDomain)

    return {'version': '0.1.1', 'parallel_read_safe': True}



"""
    antidox.doxy
    ~~~~~~~~~~~~

    Parse Doxygen XML Files.

    .. admonition: note
        Not everything is supported: some kind of compounds and some
        attributes will be ignored.
"""
# TODO: add a better overview

import os
from io import StringIO
import re
import enum
import sqlite3
from collections import namedtuple
import itertools
import pathlib
import functools

from lxml import etree as ET

__author__ = "Juan I Carrano"
__copyright__ = "Copyright 2018, Freie Universität Berlin"

# TODO: proper logging of warnings


@enum.unique
class Kind(enum.Enum):
    """Combination of Doxygen's "CompoundKind" and "MemberKind".
    Only the kinds that make sense for C and C++ are included.
    """
    CLASS = 0
    STRUCT = 1
    UNION = 2
    EXCEPTION = 3
    FILE = 4
    NAMESPACE = 5
    GROUP = 6
    PAGE = 7
    EXAMPLE = 8
    DIR = 9
    DEFINE = 10
    PROPERTY = 11
    VARIABLE = 12
    TYPEDEF = 13
    ENUM = 14
    ENUMVALUE = 15
    FUNCTION = 16
    FRIEND = 17

    # TODO: factor this out into a superclass
    def __conform__(self, protocol):
        if protocol is sqlite3.PrepareProtocol:
            return str(self.value)

    @classmethod
    def compounds(cls):
        """Return a tuple containing all kinds that are compounds.

        While we generally disregard doxygen categories in this module, the
        member/compound definition is relevant because compounds have their
        own files.
        """
        return (cls.CLASS, cls.STRUCT, cls.UNION, cls.EXCEPTION,
                cls.FILE, cls.NAMESPACE, cls.GROUP, cls.PAGE, cls.DIR)

    @classmethod
    def synthetic_compounds(cls):
        """Return a tuple containing all kinds that are compounds and are defined
        by the user and not by the language syntax"""

        return (cls.GROUP, cls.PAGE, cls.DIR)

    # FIXME: add  instance methods (like "is_subordinate")

    @classmethod
    def subordinate(cls):
        """Return a tuple containing those kinds that are not proper members
        (i.e. they are not defined by memberdef) but rather are "children" of
        a member."""
        return (cls.ENUMVALUE,)

    @classmethod
    def tag_supported(cls, attr):
        """Check if we support a xml "kind" attribute"""
        return attr.upper() in cls.__members__

    @classmethod
    def from_attr(cls, attr):
        """Convert a xml "kind" attribute to a Kind object"""
        try:
            return cls[attr.upper()]
        except KeyError as e:
            # TODO: catch this exception and issue a warning
            raise NotImplementedError("kind=%s not supported" % attr) from e


sqlite3.register_converter("Kind", lambda x: Kind(int(x)))


def _ez_iterparse(filename, events=()):
    """Wrapper around ElementTree.iterparse() that clears away elements after
    they are used, thus freeing memory.

    Note that when the "end" event is emmited for a parent, it's children will
    have already been cleared.

    To make sure the elements are garbage collected, do not keep any references.
    """

    _events = tuple(set(events + ("end",)))

    for event, elem in ET.iterparse(filename, events=_events):
        if event in events:
            yield event, elem

        if event == "end":
            # see https://lxml.de/parsing.html#modifying-the-tree
            elem.clear()
            if elem.getprevious() is not None:
                elem.getparent()[0]


class DoxyFormatError(Exception):
    """Error for wrongly formatted doxygen files"""
    pass


class ConsistencyError(DoxyFormatError):
    """Raised when any of the assumptions made about the structure of the
    Doxygen XML files is violated. For example, if more than one file contains
    an element.
    This exception is meant to replace assertions. It makes it easier to
    understand what went wrong.
    """
    pass


class RefError(Exception):
    """Base class for errors related to refids and targets"""
    pass


class InvalidTarget(RefError):
    """Raised when an invalid target (one that resolved to zero entities) is
    encountered"""
    pass


class AmbiguousTarget(RefError):
    """Raised when a target that matches more than one entity is encountered.
    The second argument to the constructor should be a list of all matches."""
    pass


# Reverse engineered Doxygen refid:
#   "_" is a escape character. A literal "_" is represented by "__". "_1" is
#   the separator we are looking for, and represents ":".
#   refids are a sort of namespaced identifier. (:) separates components.
#   We will split it into an (optional) prefix, and an id, where the id cannot
#   contain ":"
_refid_re = re.compile(r"(?:((?:\w|-)+)_1)?((?:(?:[A-Za-z0-9-]+)|(?:_[^1]))+)")


_RefId = namedtuple("_RefId", "prefix id_")


class RefId(_RefId):
    """Reverse engineered Doxygen refid.

    refids are a sort of namespaced identifier. (:) separates components.
    We will split it into an (optional) prefix, and an id, where the id cannot
    contain ":"

    "_" is a escape character. A literal "_" is represented by "__". "_1" is
    the separator we are looking for, and represents ":".

    This object can be constructed in two ways:

    * From a string.
    * From separate ``prefix`` and ``id_`` components.
    """

    def __new__(cls, *args, **kwargs):
        if len(args) == 1 and not kwargs:
            s = args[0]

            if isinstance(s, cls):
                return s

            match = _refid_re.fullmatch(s)
            if not match:
                raise DoxyFormatError("Cannot parse refid: %s" % s)

            p, h = match.groups()

            return super().__new__(cls, p or "", h)
        else:
            return super().__new__(cls, *args, **kwargs)

    def __str__(self):
        prefix, id_ = self
        return "{}_1{}".format(prefix, id_) if prefix else id_


# String of the form  [[<dir>/]*<file>::][ns::]name
#                     <--- File path -> <---name-->
# Note that we allow everything in the name. The string for a file consists
# of an empty file path and the filename as the name.
_target_re = re.compile(r"(?:((?:[^/]+/)*[^/]+\.[^/:.]+)::)?(.+)")


_Target = namedtuple("_Target", "path name")


# On a test run, the following optimization alone cut execution time of
# sphinx-build from 3'30'' to 2'55'. This makes the total time be dominated
# by the writing step, which does not depend on this extension.
@functools.lru_cache(maxsize=32)
def _parse_xml(filename):
    """Parse a xml file into an ElementTree. This function is cached for
    performance since during normal use the same file is frequently accessed
    many times in a row."""
    with open(filename) as f:
        return ET.parse(f)


class Target(_Target):
    """Tuple uniquely identifying an entity.

    In contrast to a refid, a target string can be reasonably derived by a
    human by reading the source code.
    A target string of the form "some/dir/components/file::entity", with
    "some/dir/components" as long as necessary to make the file unique.
    The entity may be namespaced with "::" like in C++ if it is defined
    inside another entity (e.g. a struct member).

    This class represents the string as a tuple.

    This object can be constructed in two ways:

    * From a string.
    * From separate path and name components. Additionally, the name component
      may be either a string or an iterable yielding name sub-components.

    A file entity is represented as ``<path>::*`` (i.e., the name filed is
    ``*``).
    """
    def __new__(cls, *args):
        # FIXME: this does not match namedtuple's __new__ signature

        if len(args) == 1:
            if isinstance(args[0], cls):
                return args[0]

            match = _target_re.fullmatch(args[0])
            if not match:
                raise ValueError("Malformed target string: %s" % args[0])
            path, name = match.groups()
        else:
            path, name = args
            if not isinstance(name, str):
                name = "::".join(name)

        return super().__new__(cls, path, name)

    def __str__(self):
        return "{}::{}".format(*self) if self.path else self.name

    @property
    def name_components(self):
        return self.name.split('::')


SearchResult = namedtuple("SearchResult", "refid name kind")
"""Container for the result of find_children and find_parents queries."""


def _match_path(p1, p2):
    """Compare two paths from right to left and return True if they could refer
    to the same file.

    As a special case, if the second argument is None, or empty, it is always
    considered a match. This simplifies query logic when the target does not
    have a path component.

    If p2 starts with "./" then the paths must math entirely. This to allow
    addressing in the case where a path is a prefix of another.
    """
    if not p2:
        return True

    part1 = pathlib.Path(p1).parts
    part2 = pathlib.Path(p2).parts

    if p2.startswith(".") and part2 and not part2[0].startswith("."):
        minlen = 0
    else:
        minlen = min(len(part1), len(part2))

    return part1[-minlen:] == part2[-minlen:]


def _barename(n):
    """Strip the namespace part of a name."""
    return n.split('::')[-1]


def _refid_str(f):
    """Decorator to make a function that accepts a refid also accept the string"""

    @functools.wraps(f)
    def _f(self, refid, *args, **kwargs):
        return f(self, RefId(refid), *args, **kwargs)

    return _f


def _target_str(f):
    """Decorator to make a function that accepts a target also accept the string"""

    @functools.wraps(f)
    def _f(self, target, *args, **kwargs):
        return f(self, Target(target), *args, **kwargs)

    return _f


class DoxyDB:
    """Interface to the Doxygen DB

    The Doxygen DB is just a directory filled with xml files. It should contain
    an "index.xml".

    Doxygen contains compounds and members. We will refer to both as "elements".
    The nesting of elements seems quite arbitrary, things like "function" can
    appear nested under both a "file" and a "group". "struct" in the other hand,
    appear as top-level in the index, though in reality they are contained in a
    file and in maybe a group.

    DoxyDB uses a SQLite database to sort this problem. Interestingly, doxygen
    can create a sqlite3 db, but it's not very well documented (ironic, isn't
    it?).

    Read-only index: After the initial database creation, no further modification are done by any
    method. This ensures DoxyDB is safe to use for parallel builds (where there
    will be multiple independent processes, each with a copy of the in-memory DB)

    refid: Each element in Doxygen is uniquely defined by a "refid", consisting of a
    string of the form string_part_1id_part.
    """
    # TODO: check if a file can be used (and shared) instead if ":memory:"

    def __init__(self, xml_dir):
        self._xml_dir = xml_dir
        self._db_conn = None

        self._init_db()

        self._read_index(os.path.join(self._xml_dir, "index.xml"))
        self._load_all_inner()
        self._vacuum()

    # Pickle support
    def __getstate__(self):
        """Dump the database as SQL commands so that it can be pickled."""
        db_dump = StringIO()

        # It should be safe to assume that the DB is connected (because it is
        # done in __init__
        db_dump.writelines(self._db_conn.iterdump())

        return {'_xml_dir': self._xml_dir, '_db_dump': db_dump.getvalue()}

    def __setstate__(self, state):
        self._xml_dir = state['_xml_dir']
        self._db_conn = None
        self._create_db_conn()
        self._db_conn.executescript(state['_db_dump'])
        self._vacuum()

    def _vacuum(self):
        old_isolation = self._db_conn.isolation_level
        self._db_conn.isolation_level = None
        try:
            self._db_conn.execute("VACUUM")
        finally:
            self._db_conn.isolation_level = old_isolation

    def _create_db_conn(self):
        """Initialize the DB connection and configure it."""

        if self._db_conn is not None:
            self._db_conn.close()
            self._db_conn = None

        # TODO: investigate the benefits of using an actual file for the DB
        self._db_conn = sqlite3.connect(':memory:',
                                        detect_types=sqlite3.PARSE_DECLTYPES)

        self._db_conn.row_factory = sqlite3.Row
        self._db_conn.create_function("match_path", 2, _match_path)
        self._db_conn.create_function("barename", 1, _barename)

    def _init_db(self):
        """Create a DB in memory and create empty tables."""
        self._create_db_conn()

        # Example:
        # <compound refid="fxos8700__regs_8h" kind="file"><name>fxos8700_regs.h</name>
        #   <member refid="fxos8700__regs_8h_1abd2eb1f9d6401758c261450bf6f78280" kind="define">
        #        <name>FXOS8700_REG_STATUS</name>
        #   </member>
        #  ....
        # An entry is added to the elements table:
        #   prefix=fxos8700__regs_8h, id=abd2eb1f9d6401758c261450bf6f78280,
        #   kind="define", name="FXOS8700_REG_STATUS"
        # And an entry will be added to the hierarchy table
        #   prefix=fxos8700__regs_8h, id=abd2eb1f9d6401758c261450bf6f78280,
        #   p_prefix="", p_id="fxos8700__regs_8h"
        #
        self._db_conn.executescript("""
        PRAGMA foreign_keys = 1;

        CREATE TABLE elements (
            prefix TEXT, id TEXT,
            name TEXT NOT NULL,
            kind Kind NOT NULL,
            PRIMARY KEY (prefix, id) ON CONFLICT IGNORE
            );

        CREATE TABLE hierarchy (
            prefix TEXT NOT NULL, id TEXT NOT NULL,
            p_prefix TEXT NOT NULL, p_id TEXT NOT NULL,
            UNIQUE (prefix, id, p_prefix, p_id) ON CONFLICT REPLACE,
            FOREIGN KEY(prefix, id) REFERENCES elements(prefix, id)
            );

        CREATE TABLE compound_kinds (kind Kind NOT NULL,
                                     UNIQUE(kind)
                                     );
        CREATE TABLE syn_compound_kinds (kind Kind NOT NULL,
                                         UNIQUE(kind)
                                         );
        """)

        _compounds = Kind.compounds()
        self._db_conn.executemany("INSERT INTO compound_kinds VALUES (?)",
                                  ((x,) for x in _compounds))

        _syn_compounds = Kind.synthetic_compounds()
        self._db_conn.executemany("INSERT INTO syn_compound_kinds VALUES (?)",
                                  ((x,) for x in _syn_compounds))

    def _insert_element(self, refid, name, kind, parent_refid=None):
        """Insert an element into the database."""

        try:
            self._db_conn.execute("INSERT INTO elements values "
                                  "(?, ?, ?, ?)",
                                  refid + (name, kind))
        except sqlite3.IntegrityError:
            print(refid, name, kind)  # FIXME: replace by proper logging
            raise

        if parent_refid is not None:
            self._db_conn.execute("INSERT INTO hierarchy values (?, ?, ?, ?)",
                                  refid + parent_refid)

    def _read_index(self, indexfile):
        """Parse index.xml and insert the elements in the database."""
        for event, elem in _ez_iterparse(indexfile, ("end",)):
            if elem.tag == "doxygenindex":
                continue

            # Doxygen puts the name of an element in a child element <name>
            # instead of an attribute. Here we move it so that it is like we
            # want it to be, and of course we skip the <name> element.
            if elem.tag == "name":
                elem.getparent().attrib["name"] = elem.text
                continue

            if elem.tag == "compound":
                p_refid = None
            elif elem.tag == "member":
                p_refid = RefId(elem.getparent().attrib["refid"])
            else:
                raise DoxyFormatError("Unknown tag in index: %s"
                                      % elem.tag)

            this_refid = RefId(elem.attrib["refid"])
            kind = Kind.from_attr(elem.attrib["kind"])

            # Doxygen wrongly places enumvalues as direct children of files
            # instead of the containing enum.
            # Let's skip the parent for now and we set it correctly when
            # reading the inner files.
            if kind == Kind.ENUMVALUE:
                p_refid = None

            try:
                name = elem.attrib["name"]
            except KeyError as e:
                raise DoxyFormatError("Element definition without a name: %s"
                                      % elem.attrib["refid"]) from e

            self._insert_element(this_refid, name, kind, p_refid)


    def _load_all_inner(self):
        """Load the XML file for each compound and assemble the hierarchy."""
        cur = self._db_conn.execute(
                "SELECT prefix, id FROM elements WHERE kind in compound_kinds")

        for refid in cur:
            fn = os.path.join(self._xml_dir, "{}.xml".format(RefId(*refid)))
            self._read_inner(fn)

    def _read_inner(self, compoundfile):
        """Gather all the inner elements for compounds in a file."""

        for event, elem in _ez_iterparse(compoundfile, ("start",)):
            if elem.tag == "doxygen":
                continue

            if elem.tag == "compounddef":
                p_refid = RefId(elem.attrib["id"])
            else:
                # the enumvalue is a workaround to nest enumvalues under enums
                if elem.tag == "enumvalue":
                    parent_elem = elem.getparent()
                    if not parent_elem.tag == "memberdef":
                        raise ConsistencyError(
                            "expected parent of enumvalue to be a memberdef")
                    this_parent = RefId(parent_elem.attrib["id"])
                    id_attr = elem.attrib["id"]
                else:
                    s, inner, innerkind = elem.tag.partition("inner")
                    if s:  # the tag does not start with "inner"
                        continue

                    if not Kind.tag_supported(innerkind):
                        continue

                    this_parent = p_refid
                    id_attr = elem.attrib["refid"]

                this_refid = RefId(id_attr)

                self._db_conn.execute("INSERT INTO hierarchy values (?, ?, ?, ?)",
                                      this_refid + this_parent)

    # TODO: this may need caching???
    @_refid_str
    def find_parents(self, refid):
        """Get the refid of the compounds where the given element is defined.

        Parameters
        ----------

        refid : refid or str
            Doxygen "id" field for the element.

        Returns
        -------

        results: iterable yielding SearchResult
            All direct ancestors of this element.
        """
        cur = self._db_conn.execute(
        """SELECT p_prefix, p_id, name, kind
        FROM hierarchy INNER JOIN elements
            ON hierarchy.p_prefix = elements.prefix AND hierarchy.p_id = elements.id
        WHERE hierarchy.prefix = ?
              AND hierarchy.id = ?
              AND kind in compound_kinds""", refid)

        return (SearchResult(RefId(*ref), name, kind) for *ref, name, kind in cur)

    @_refid_str
    def find_children(self, refid):
        """Find all members and compounds that are a direct descendants of this
        element.

        Returns
        -------

        members: list of SearchResult
            Descendents that cannot contain any children themselves.
        compounds: list of SearchResult
            Descendents that are compounds, and as such may contain children.
        """
        # TODO: add parameter to filter by kind
        cur = self._db_conn.execute(
        """SELECT hierarchy.prefix, hierarchy.id, name, kind,
                  kind IN compound_kinds as is_compound
        FROM hierarchy INNER JOIN elements
            ON hierarchy.prefix = elements.prefix AND hierarchy.id = elements.id
        WHERE hierarchy.p_prefix = ?
              AND hierarchy.p_id = ?
        ORDER BY
              is_compound""",
            refid)

        r = [(), ()]
        for iscompound, g in itertools.groupby(cur, lambda x: x["is_compound"]):
            r[iscompound] = [SearchResult(RefId(prefix, _id), name, kind)
                             for prefix, _id, name, kind, _ in g]

        return r

    def find(self, kinds = None, no_parent = False):
        """Find all elements of the specified kinds.

        Parameters
        ----------

        kinds: list of Kind to filter by. If not give, all kinds are retrieved
            (except those listed in Kind.subordinate())
        no_parent: if True, return only elements without a parent.

        Returns
        -------

        result: iterable yielding SearchResult
        """

        if kinds is not None:
            _kinds = kinds
        else:
            _kinds = list(set(Kind.__members__.values())
                          - set(Kind.subordinate()))

        query = (
        """WITH
            allowed_kinds (kind) AS (
                VALUES {}
            )
            SELECT elements.* FROM elements
            INNER JOIN allowed_kinds as ak
                ON ak.kind == elements.kind
        """).format(",".join(itertools.repeat("(?)", len(_kinds))))

        if no_parent:
            query += """
            LEFT JOIN hierarchy as h
                ON h.prefix == elements.prefix AND h.id == elements.id
            WHERE h.p_prefix IS NULL AND h.p_id IS NULL
            """

        cur = self._db_conn.execute(query, _kinds)

        return (SearchResult(RefId(*ref), name, kind)
                for *ref, name, kind in cur)

    @_refid_str
    def refid_to_target(self, refid):
        """Generate a target tuple uniquely identifying a refid.

        Since targets must be descendents if a file element, this method will
        fail for user-defined constructs like groups.
        """
        # If we omit user-defined constructs like groups, the elements form
        # a tree, where the files are roots.
        # The query below traverses the tree until it reaches a file and returns
        # the nodes.
        cur = self._db_conn.execute(
        """WITH RECURSIVE
            parent (level, prefix, id) AS (
                SELECT 0, prefix, id FROM elements
                    WHERE prefix = ? AND id = ?
                UNION
                SELECT parent.level + 1, elements.prefix, elements.id
                FROM parent
                    INNER JOIN hierarchy
                        ON hierarchy.prefix = parent.prefix AND
                           hierarchy.id = parent.id
                    INNER JOIN elements
                        ON hierarchy.p_prefix = elements.prefix AND
                           hierarchy.p_id = elements.id
                WHERE NOT elements.kind IN syn_compound_kinds
                LIMIT 20
            )
        SELECT name, kind FROM elements INNER JOIN parent
            ON elements.prefix = parent.prefix AND elements.id = parent.id
        GROUP BY level

        """, refid)

        nodes = list(cur)

        if not len(nodes) > 0:
            raise InvalidTarget("No such refid: %s" % str(refid))

        if not nodes[-1]['kind'] == Kind.FILE:
            raise ConsistencyError("Root node is not a file")

        if len(nodes) == 1:
            # Fix for #15. Check if the resulting name is ambiguous.
            path = nodes[0]['name']
            n_matching = self._db_conn.execute(
                """SELECT COUNT(*) FROM elements
                    WHERE kind = ? AND  match_path(name, ?)
                """, (Kind.FILE, path)).fetchone()[0]

            return Target(path if n_matching == 1
                          else ".{}{}".format(os.path.sep, path), '*')
        else:
            return Target(nodes[-1]['name'],
                          (n['name'] for n in reversed(nodes[:-1])))

    @_refid_str
    def get(self, refid):
        """Get the Name, Kind for of an element.

        Returns
        -------

        row object (similar to a named tuple)
            with fields "name", "kind".
        """
        cur = self._db_conn.execute(
            """SELECT name, kind FROM elements WHERE prefix = ? AND id = ?""",
            refid)

        # No need to check for more than one result, prefix and id are primary keys
        try:
            result = list(cur)[0]
        except IndexError as e:
            raise RefError("No such refid: %s" % str(refid)) from e

        return result

    @staticmethod
    def _cur_to_refid(cur, target, scoped = False):
        """Turn a cursor into a Refid. Raise errors if it is empty or has more
        than one element (except if scoped is True, see below.)

        The cursor should contain 3 columns, and be ordered in descending order
        according to the first one:

        in_scope
            true if the target is a direct descendant of a "scope"
            parameter. This is only relevant if scoped = True.
        prefix
            prefix of refid
        id
            id of refid

        Multiple results are tolerated if scoped is True, and there is exactly
        one result with in_scope = True.
        """

        r = list(cur)

        scoped_results = (x for x in r if x["in_scope"])

        if not r:
            raise InvalidTarget("Cannot resolve target: %s" % str(target))
        # FIXME: this is failing for paths that are a prefix of another one.
        if len(r) > 1 and (not scoped or not r[0]["in_scope"]
                           or len(list(scoped_results)) > 1):
            raise AmbiguousTarget("Target (%s) resolves to more than one element"
                                      % str(target), [RefId(*row[1:]) for row in r])

        return RefId(*r[0][1:])

    @_target_str
    def resolve_target(self, target, scope = None):
        """Convert a target string into a refid.

        This method accepts a string of the form
            [[<dir>/]*<file>::][ns::name]

        If scope is given it must be a RefId or refid-compatible string.

        The path and the amount of directory components included is optional as
        long as it resolves univocally.
        If the string is ambiguous (i.e., more than one entity matches, an error
        is raised).
        The scope parameter allows for disambiguation by preferring results that
        are children of a given refid. Even then, if there is more than one
        result that matches both conditions, it is still an error.
        Because of the way Doxygen works with C, if there is a namespace it
        must be specified. This only happens with structs/unions defined inside
        other struct/unions.
        """
        components = tuple(target.name_components)
        ncompo = len(components)

        # Accept matches at level zero if the target refers to a file.
        if ncompo == 1 and target.name == '*':
            accept_level = 0
        else:
            accept_level = ncompo

        path_filter = target.path

        scope_prefix, scope_id = RefId(scope) if scope else ("", "")

        # The call to barename is a kind of hack. It is necessary because
        #      doxygen stores some names with namespaces and some without.
        # The DISTINCT keyword is there because sometimes the search returns
        #      the same entity multiple times. I think it may only be because of
        #      some bug in doxygen (further investigation is needed.)
        cur = self._db_conn.execute(
        """WITH RECURSIVE
            components (level, compo) AS (
                VALUES %s
            ),
            follow (level, prefix, id) AS (
                SELECT 0, prefix, id FROM elements
                    WHERE kind = ? AND match_path(name, ?)
                UNION ALL
                SELECT f.level + 1, h.prefix, h.id
                FROM follow AS f
                    INNER JOIN hierarchy AS h
                        ON h.p_prefix = f.prefix AND h.p_id = f.id
                    INNER JOIN elements AS e
                        ON h.prefix = e.prefix AND h.id = e.id
                    INNER JOIN components AS c
                        ON f.level = c.level
                WHERE barename(e.name) = c.compo
            )
        SELECT MAX(h.p_prefix = ? AND h.p_id = ?) AS in_scope, f.prefix, f.id FROM
            (SELECT DISTINCT prefix, id FROM follow
                 WHERE level = ?
            ) AS f
            LEFT JOIN hierarchy AS h
                ON h.prefix = f.prefix AND h.id = f.id
        GROUP BY f.prefix, f.id
        ORDER BY in_scope DESC
        """ % ",".join("(%s, ?)" % i for i in range(ncompo)),
            components + (Kind.FILE, path_filter, scope_prefix, scope_id,
                          accept_level))

        return self._cur_to_refid(cur, target, scope is not None)

    def resolve_name(self, kind, name, scope = None):
        """Find an element with the specified kind and name. If kind is not given,
        all kinds are searched.

        More than one result will trigger an AmbiguousTarget exception.
        Less than one, and InvalidTarget.

        Ambiguity can be saved by providing a scope similar to `resolve_target`.
        """

        scope_prefix, scope_id = RefId(scope) if scope else ("", "")

        cur = self._db_conn.execute(
        """SELECT MAX(h.p_prefix = :scope_prefix AND h.p_id = :scope_id)
                AS in_scope, e.prefix AS prefix, e.id AS id
        FROM elements as e LEFT JOIN hierarchy as h
          ON h.prefix = e.prefix AND h.id = e.id
        WHERE (:ignore_kind OR e.kind = :kind) AND e.name = :name
        GROUP BY e.prefix, e.id
        ORDER BY in_scope DESC
        """,
        {"scope_prefix": scope_prefix, "scope_id": scope_id,
        "ignore_kind": not bool(kind), "kind": kind, "name": name})

        return self._cur_to_refid(cur, (kind, name), scope is not None)

    def _first_parent(self, refid, kind):
        """Return the first compound containing an entity."""
        if kind in Kind.subordinate():
            cur = self._db_conn.execute(
            """SELECT h1.p_prefix, h1.p_id
            FROM hierarchy as h1 INNER JOIN elements
                ON h1.p_prefix = elements.prefix AND h1.p_id = elements.id
            INNER JOIN hierarchy as h2
                ON h1.prefix = h2.p_prefix AND h1.id = h2.p_id
            WHERE h2.prefix = ?
                AND h2.id = ?
                AND kind in compound_kinds
            """, refid)
            solutions = (RefId(*ref) for ref in cur)
        else:
            solutions = (res.refid for res in self.find_parents(refid))

        # Doxygen defines the same member in more than one place.
        # It is wasteful, though it makes our life easier.
        try:
            return next(solutions)
        except StopIteration as e:
            raise ConsistencyError(
                    "Cannot find compound containing {}".format(refid)) from e

    @_refid_str
    def get_tree(self, refid):
        """Get the xml element tree for an element"""

        refkind = self.get(refid)['kind']
        if refkind in Kind.compounds():
            # compounds are defined in their own file.
            definition_file_base = refid
            xpathq = '//compounddef[@id=$id]'
        else:
            definition_file_base = self._first_parent(refid, refkind)

            # FIXME: this code looks ugly
            xpathq = ('//memberdef[@id=$id]'
                      if not refkind in Kind.subordinate()
                      else '//{}[@id=$id]'.format(refkind.name.lower()))

        # TODO: should we cache this?
        fn = os.path.join(self._xml_dir, "{}.xml".format(definition_file_base))
        compound_doc = _parse_xml(fn)

        return compound_doc.xpath(xpathq, id=str(refid))[0]

    # TODO: these should be supported. How?
    # For these ones, maybe have an indexing directive
    #   doxy.Kind.FILE = 4
    #   doxy.Kind.NAMESPACE = 5
    #   doxy.Kind.GROUP = 6
    #   doxy.Kind.EXAMPLE = 8
    # For these other ones maybe add new C roles
    #   doxy.Kind.PROPERTY = 11
    #   doxy.Kind.VARIABLE = 12
    #   doxy.Kind.ENUM = 14
    #   doxy.Kind.ENUMVALUE = 15
    # These ones we can ignore
    #   doxy.Kind.FRIEND = 17
    #   doxy.Kind.PAGE = 7
    #   doxy.Kind.DIR = 9

    _easy_kinds = {
        Kind.ENUM: 'type',
        Kind.STRUCT: 'type',
        Kind.UNION: 'type',
        Kind.TYPEDEF: 'type',
        Kind.DEFINE: 'macro',
        Kind.FUNCTION: 'function'
        }

    def guess_desctype(self, refid):
        """Try to guess the "real" type (a C domain role) of a doxygen element.

        This is somewhat related to the Kind, but also depends on the element's parent.
        """

        name, kind = self.get(refid)
        if kind in self._easy_kinds:
            return self._easy_kinds[kind]

        if kind == Kind.VARIABLE:
            # Doxygen uses "variable" for structure/class members and for
            # actual variables.
            # TODO: do this with a query
            _struct_like = (Kind.UNION, Kind.STRUCT)
            parent_is_struct = any(res.kind in _struct_like
                                   for res in self.find_parents(refid))

            return "member" if parent_is_struct else "var"

        # FIXME: is this correct?
        raise ValueError("No c desctype for %s" % kind)

    # TODO: hierarchy walker (sort of os.walkdir with compounds as dirs and members
    #       as files???????)



from setuptools import setup


def readme():
    with open('README.rst') as f:
        return f.read()

setup(name='antidox',
      version='0.1.4',
      description='Sphinx extension to extract and insert Doxygen documentation.',
      long_description=readme(),
      long_description_content_type='text/x-rst',
      classifiers=[
        'Development Status :: 4 - Beta',
        'License :: OSI Approved :: BSD License',
        'Programming Language :: Python :: 3',
        'Intended Audience :: Developers',
        'Framework :: Sphinx :: Extension',
        'Topic :: Documentation :: Sphinx',
        'Topic :: Software Development :: Documentation'
      ],
      keywords='sphinx doxygen',
      url='https://github.com/jcarrano/antidox',
      project_urls={
        'Documentation': 'https://antidox.readthedocs.io/en/latest/index.html',
        'Tracker': 'https://github.com/jcarrano/antidox/issues',
      },
      author='Juan I Carrano <juan@carrano.com.ar>',
      author_email='juan@carrano.com.ar',
      license='BSD',
      packages=['antidox'],
      install_requires=[
        'sphinx',
        'lxml'
      ],
      entry_points={
        'console_scripts': [
            'antidox-shell = antidox.shell:main',
        ],
      },
      include_package_data=True,
      zip_safe=True
)



"""
    antidox.xtransform
    ~~~~~~~~~~~~~~~~~~

    XSLT handling code. This code is placed here so that is can be used without
    importing the sphinx/rest stuff.
"""

import os
import collections
import re
import functools
from pkgutil import get_data

from lxml import etree as ET

__author__ = "Juan I Carrano"
__copyright__ = "Copyright 2018, Freie Universität Berlin"


# Load basic text functions


def _to_text(nodes_or_text):
    """Convert a xpath object (a string or a list of nodes) to string"""
    element0 = (nodes_or_text[0] if isinstance(nodes_or_text, list)
                else nodes_or_text)
    return (element0 if isinstance(element0, str)
            else element0.xpath("string(.)"))


def _textfunc(f):
    """Decorator to convert the arguments of an XPath extension function to
    string."""
    @functools.wraps(f)
    def _f(ctx, nodes_or_text):
        return f(_to_text(nodes_or_text))

    return _f


_basic_ns = ET.FunctionNamespace("antidox")


@_basic_ns("upper-case")
@_textfunc
def _upper_case(text):
    return text.upper()


@_basic_ns("lower-case")
@_textfunc
def _lower_case(text):
    return text.lower()


@_basic_ns("string-to-ids")
@_textfunc
def _string_to_ids(text):
    """Convert a string into something that is safe to use as a docutils ids
    field."""
    return _NORMALIZE_SPACE.sub(
        "-", text.strip()).translate(_XLATE_TABLE)


def _get_compound_xsl_text():
    return get_data(__package__, os.path.join("templates", "compound.xsl"))


class Resolver(ET.Resolver):
    """Resolve the basic stylesheet as "antidox:compound".

    If this package is installed as a zip, the XML may not even be a file, so
    we hide those details from the user."""

    def resolve(self, url, id, context):
        if url == "antidox:compound":
            return self.resolve_string(_get_compound_xsl_text(), context)


_az = range(ord('a'), ord('z'))
_09 = range(ord('0'), ord('9'))
_dash = [ord("-"), ord("_")]

_XLATE_TABLE = collections.defaultdict(
    lambda: None,
    list(zip(_az, _az)) + list(zip(_09, _09))
    + [(ord(chr(x).upper()), x) for x in _az]
    + list(zip(_dash, _dash)) + [(ord(" "), ord("-"))])


_NORMALIZE_SPACE = re.compile(" +")


def _textmeth(f):
    """Decorator to convert the arguments of an XPath extension method to
    string."""
    @functools.wraps(f)
    def _f(self, ctx, nodes_or_text):
        return f(self, ctx, _to_text(nodes_or_text))

    return _f


# From https://stackoverflow.com/a/33527584/5220128
# Split commas.
_SPLIT_COMMA = re.compile(r',(?!(?:[^(]*\([^)]*\))*[^()]*\))')


class _XPathExtensions:
    def __init__(self, locale_fn=None, doxy_db=None):
        self._locale_fn = locale_fn or (lambda x: x)
        self._doxy_db = doxy_db

    @_textmeth
    def l(self, _, text):
        """Stub locale function. This is here so we can run the XSL transform
        without having to import sphinx."""
        return self._locale_fn(text)

    @_textmeth
    def guess_desctype(self, _, text):
        return "" if not self._doxy_db else self._doxy_db.guess_desctype(text)

    @_textmeth
    def refid_to_target(self, _, text):
        return "" if not self._doxy_db else str(self._doxy_db.refid_to_target(text))

    @_textmeth
    def parse_argstr(self, _, text):
        """Convert an argument string into desc_parameter nodes. This is used
        to work around a bug in Doxygen where it will not parse arguments to
        function pointers."""
        nodes = []
        for arg in _SPLIT_COMMA.split(text):
            atype, _s, aname = arg.rpartition(' ')
            desc_p = ET.Element('desc_parameter')
            desc_t = ET.SubElement(desc_p, 'desc_type')
            desc_t.text = atype
            if aname:
                desc_t.tail = " "
                ET.SubElement(desc_p, 'desc_name').text = aname
            nodes.append(desc_p)

        return nodes


def get_stylesheet(stylesheet_filename=None, locale_fn=None, doxy_db=None):
    """Get a XSLT stylesheet.

    If stylesheet_filename is not specified, the default sheet will be returned.

    If a file is given, it will be loaded with a special loader that exposes
    the default one under the "antidox:compound".

    Parameters
    ----------
    stylesheet_filename: XSL style sheet, or None to load the default value.
    locale_fn: A translation function mapping strings to strings. This will be
               to implement the "antidox:l" XPath function. If not given,
               the identity function will be used.
    """

    custom_extension = _XPathExtensions(locale_fn, doxy_db)

    ext = ET.Extension(custom_extension, ns="antidox")

    if stylesheet_filename:
        parser = ET.XMLParser()
        parser.resolvers.add(Resolver())
        xml_doc = ET.parse(stylesheet_filename, parser)
    else:
        xml_doc = ET.XML(_get_compound_xsl_text())

    return ET.XSLT(xml_doc, extensions=ext)



"""
    antidox.shell
    ~~~~~~~~~~~~~

    Simple debug shell. This application lets you query the database that
    is generated from Doxygen XML docs.

    Access it by typing::

      python -m antidox.shell

    or by using the ``antidox-shell`` console script.
"""


import cmd
import argparse
import functools
import pickle
import sqlite3
import timeit
import re
import pathlib

from lxml import etree as ET

from . import doxy
from .xtransform import get_stylesheet

__author__ = "Juan I Carrano"
__copyright__ = "Copyright 2018, Freie Universität Berlin"


def _catch(*args):
    """Wrap a function so that it catches errors"""
    exceptions = args or Exception

    def _catch_decorator(f):
        @functools.wraps(f)
        def _f(self, line):
            try:
                return f(self, line)
            except exceptions as e:
                print(" ".join(str(a) for a in e.args))

        return _f

    return _catch_decorator

_catch_doxy = _catch(doxy.RefError, doxy.DoxyFormatError)

def _any_to_refid(f):
    """Transform a method that takes a refid into one that takes a
    character "r" or "t" and either a refid or a target."""
    @functools.wraps(f)
    def _f(self, line):
        try:
            type, refid_or_target, *extra = line.split()
        except ValueError:
            print("Error: expected at least two arguments.")
            return

        if type not in ("r", "t"):
            print("Usage:")
            print(f.__doc__)
            return

        refid = (refid_or_target if type == 'r'
                 else self.db.resolve_target(refid_or_target))

        return f(self, refid, *extra)

    return _f


KIND_RE = re.compile(r'(?<=\s)Kind\.([A-Z]+)')


class Shell(cmd.Cmd):
    """Interact with the database created by that antidox.doxy module."""

    prompt = "antidox> "
    intro = 'This is the antidox debug shell. Type "?" or "help" for help'

    """Command that can be run withour a database loaded"""
    NOINIT_CMDS = ("", "info", "new", "restore", "load_sphinx", "?", "!", "EOF",
                   "help", "sty")

    def __init__(self, doxydb=None, **kwargs):
        self._stylesheet_fn = None
        self.db = doxydb

        super().__init__(**kwargs)

    @property
    def db(self):
        return self._db

    @db.setter
    def db(self, value):
        self._db = value
        self._reload_sty()

    def precmd(self, line):
        if line and self.db is None and line.strip().split()[0] not in self.NOINIT_CMDS:
            print("No database loaded")
            return ""
        else:
            return line

    def emptyline(self):
        # do not repeat the last command.
        pass

    def do_EOF(self, _):
        print()
        return True

    def do_info(self, _):
        """Show information about the current database."""
        if self.db is None:
            print("No DB loaded")
            return

        print("xml dir:", self.db._xml_dir)
        print("DB tables:")
        tables = self.db._db_conn.execute("SELECT sql FROM sqlite_master WHERE type = 'table'")
        for t in tables:
            print("\t", t[0])

        self._print_cursor(self.db._db_conn.execute("""
        SELECT kind AS element_kind, COUNT(*) FROM elements GROUP BY kind
        """))

        self._print_cursor(self.db._db_conn.execute("""
        SELECT COUNT(*) AS total_elements FROM elements
        """))

        self._print_cursor(self.db._db_conn.execute("""
        SELECT COUNT(*) AS number_of_parent_elements FROM
            (SELECT DISTINCT p_prefix, p_id FROM hierarchy)
        """))

    @_catch()
    def do_new(self, xml_dir):
        """\
        new <doxy xmCmd.emptyline()l dir>
        Read an XML directory and create a database. Old DB is discarded."""
        _f = lambda: doxy.DoxyDB(xml_dir)
        print("DB loaded in %f seconds" % timeit.timeit("self.db=_f()", number=1, globals=locals()))

    @_catch()
    def do_dump(self, filename):
        """\
        dump <filename>
        Dump DB as pickle"""
        with open(filename, "wb+") as f:
            pickle.dump(self.db, f)
        print("Done")

    @_catch()
    def do_restore(self, filename):
        """\
        restore <filename>
        Restore pickled DB"""
        with open(filename, "rb") as f:
            self.db = pickle.load(f)

    @_catch()
    def do_load_sphinx(self, sphinx_builddir):
        """\
        load_sphinx <sphinx_project/_build> [project_dir]
        Load the DB from a pickled Sphinx environment.
        Usually the xml directory is stored as a relative path within the DB
        instance. The default here is to take it relative to the env's srcdir
        but that can be overriden y specifying 'project_dir'"""
        builddir, *maybe_prjdir = sphinx_builddir.split()

        import sphinx.environment

        with open(pathlib.Path(
                  sphinx_builddir, "doctrees", "environment.pickle"), 'rb') as f:
            env = sphinx.environment.BuildEnvironment.load(f)

        self.db = env.antidox_db

        base_dir = maybe_prjdir[0] if maybe_prjdir else env.srcdir
        self.db._xml_dir = pathlib.Path(base_dir, self.db._xml_dir)

    @_catch_doxy
    def do_r(self, target_and_scope):
        """\
        r <target> [<scope>]
        Get the refid of a target.
        """
        if not target_and_scope:
            print("You must provide a refid")
            return

        target, *maybe_scope = target_and_scope.split()
        scope = maybe_scope[0] if maybe_scope else None
        print(self.db.resolve_target(target, scope))

    @_catch_doxy
    def do_t(self, refid):
        """\
        t <target>
        Get a target name for a refid.
        """
        if not refid:
            print("You must provide a refid")
            return

        print(self.db.refid_to_target(refid))

    @_catch_doxy
    def do_n(self, kind_name):
        """\
        n [<kind>] <name> ["in" <scope>]
        Get a refid with a name and of a kind (optional).
        """

        args = kind_name.split()

        if len(args) == 1:
            kind_s = None
            scope = None
            name = args[0]
            in_ = "in"
        elif len(args) == 2:
            kind_s, name = args
            scope = None
            in_ = "in"
        elif len(args) == 3:
            kind_s = None
            name, in_, scope = args
        elif len(args) == 4:
            kind_s, name, in_, scope = args
        else:
            print('Too many arguments')
            return

        if in_.lower() != "in":
            print('Invalid syntax (expected "in <scope>")')
            return

        kind = doxy.Kind.from_attr(kind_s) if kind_s else None

        print(self.db.resolve_name(kind, name, scope))

    def _print_results(self, rs):
        """Pretty print a list of SearchResults"""
        for r, n, k in rs:
            print("{}\t{}\t{}".format(r, k, n))

    @_catch_doxy
    @_any_to_refid
    def do_get(self, refid):
        """\
        get r <refid>
        get t target

        Show information for an entity, given as refid or target.
        """
        name, kind = self.db.get(refid)
        print("{}\t{}\t{}".format(refid, kind, name))

    @_catch_doxy
    def do_get_all(self, params):
        """\
        get-all [noparent] <kind>*

        Retrieve all entities of the specified kinds (or all kinds if none is
        given. The "noparent" parameter will filter only those entities with no
        parent.
        """

        args = params.split()
        noparent = bool(args and args[0] == 'noparent')
        kinds_s = args[1:] if noparent else args
        kinds = [doxy.Kind.from_attr(s) for s in kinds_s] or None

        self._print_results(self.db.find(kinds, no_parent=noparent))

    @_catch_doxy
    @_any_to_refid
    def do_parents(self, refid):
        """\
        parents r <refid>
        parents t <target>

        Get all compounds that include this element.
        """
        self._print_results(self.db.find_parents(refid))

    @_catch_doxy
    @_any_to_refid
    def do_children(self, refid):
        """\
        children r <refid>
        children t <target>

        Get all elements that are direct descendants of the given one.
        """
        members, compounds = self.db.find_children(refid)
        print("#members")
        self._print_results(members)
        print("#compounds")
        self._print_results(compounds)

    @_catch_doxy
    @_any_to_refid
    def do_show(self, refid):
        """\
        show r <refid>
        show t <target>

        Show the source for a given element.

        Note: this command does some whitespace modifications so that the resulting
        XML can be pretty-printed.
        """
        root = self.db.get_tree(refid)
        for element in root.iter():
            if element.tail is not None and not element.tail.strip():
                element.tail = None

            if element.text is not None and not element.text.strip():
                element.text = None

        print(ET.tostring(root, pretty_print=True, encoding='unicode'))

    @_catch(doxy.RefError, doxy.DoxyFormatError, ET.XMLSyntaxError)
    def do_sty(self, filename):
        """\
        sty [template.xsl]

        Load a XML template file. Call with no argument to restore the default.
        The database will be reloaded each time the database is loaded.
        """

        self.stylesheet = get_stylesheet(filename, doxy_db=self.db)
        self._stylesheet_fn = filename

    def _reload_sty(self):
        """Reload the stylesheet (after a database load)"""
        self.do_sty(self._stylesheet_fn)

    @_catch_doxy
    @_any_to_refid
    def do_xform(self, refid, *flags):
        """\
        xform r refid <flags>*
        xform t target <flags>*

        Fetch an element and apply the stylesheet to it.

        Note: this command does some whitespace modifications so that the resulting
        XML can be pretty-printed.

        Flags can be any of noindex, hideloc, hidedef, hidedoc.
        """
        root = self.db.get_tree(refid)

        flags = {k: "true()" for k in flags}

        transformed = self.stylesheet(root, **flags)
        for element in transformed.iter():
            if element.tail is not None and not element.tail.strip():
                element.tail = None

        print(ET.tostring(transformed, pretty_print=True, encoding='unicode'))

    def do_shell(self, line):
        """\
        Run an arbitrary SQL query on the database.
        As a shortcut you can use "!".
        As a convenience, any string of the form "Kind.XXXXX" will be replaced
        with the numeric value for that Kind (e.g Kind.GROUP, Kind.UNION).

        e.g.: `! SELECT name, id FROM elements WHERE kind in compound_kinds`
        """
        _line = KIND_RE.sub(lambda m: str(doxy.Kind[m[1]].value), line)

        try:
            self._print_cursor(self.db._db_conn.execute(_line))
        except sqlite3.Error as e:
            print("Exception while executing SQL:")
            print(e)

    @staticmethod
    def _print_cursor(cur):
        """Pretty-print the result of a query."""
        print("#{}".format("\t".join(cn[0] for cn in cur.description)))
        for r in cur:
            print("\t".join(str(x) for x in r))


def main():
    parser = argparse.ArgumentParser(description="antidox database debugger",
                                     epilog="""\
    Use the -e option to start the application with a database already open:
    shell.py -e "new path/to/xml" OR shell.py -e "restore saved_db.pickle"
    """)

    parser.add_argument('-e', action="append",
                        help="Run command as if it was given in the terminal. "
                             "Can be specified multiple times.")

    ns = parser.parse_args()

    app = Shell()

    for cmdline in ns.e or ():
        if app.onecmd(cmdline):
            break
    else:
        app.cmdloop()


if __name__ == "__main__":
    main()



"""
    antidox.collector
    ~~~~~~~~~~~~~~~~~

    Keep track of documents that need rebuilding when the Doxygen DB is
    modified.
"""


from sphinx.environment.collectors import EnvironmentCollector

__author__ = "Juan I Carrano"
__copyright__ = "Copyright 2018, Freie Universität Berlin"


class DoxyCollector(EnvironmentCollector):
    """Collect documents that can be affected by a change in the doxygen DB.
    A custom collector is needed because note_dependency() does not work
    with directories.
    In the future this could be made smarter about what to rebuild and what not
    to.
    """
    def merge_other(self, app, env, docnames, other):
        app.env.antidox_dependencies.update(other.antidox_dependencies)

    def clear_doc(self, app, env, docname):
        app.env.antidox_dependencies.discard(docname)

    def get_outdated_docs(self, app, env, added, changed, removed):
        if not hasattr(app.env, "antidox_dependencies"):
            app.env.antidox_dependencies = set()

        return [docname for docname in app.env.antidox_dependencies
                if docname not in app.env.all_docs
                or app.env.all_docs[docname] < app.env.antidox_db_date]

    def process_doc(self, *args):
        pass

    @staticmethod
    def note_dependency(env):
        """Mark the current document as depending on the doxygen database."""
        env.antidox_dependencies.add(env.docname)
